#include <kernel/thread.h>
#include <kernel/list.h>
#include <kernel/schedule.h>
#include <arch/types.h>
#ifdef DEBUG
#include <drivers/serial/serial.h>
#include <kernel/debug.h>
#endif //DEBUG


extern void task1();


static int32u *Kern_StackAlloc(fp_t pT);			// Stack Allocation
int32u Port_StackInit(fp_t pT, int32u *);			// Stack Initialization

static void fun1(void *);					// Thread function for test








/*********************************************************************************
 * Description : Stack Allocation and TCB initialization
 *
 *
 *********************************************************************************/
void 
Kern_ThreadInit(tcb_t *newtcb,int32u tid, fp_t pT,int32u prior,int16u deadLine)
{															
  newtcb->pStack = Kern_StackAlloc(pT);
  newtcb->tid = tid;
  INIT_LIST_HEAD(&newtcb->resources);	  
  INIT_LIST_HEAD(&newtcb->memList.allocLink);	  
  INIT_LIST_HEAD(&newtcb->memList.memLink);	  
  
  /*In case scheduling policy changed, fields for other scheduling are 
	needed to set up */
  INIT_LIST_HEAD(&newtcb->threadList);	  
  newtcb->prior = prior;
  newtcb->deadLine = deadLine;

  newtcb->status = CREATED;
  return ;
//  newtcb->threadList = LIST_HEAD_INIT(newtcb->threadList);
}	




/*********************************************************************************
 * Description : Stack Allocation and TCB initialization
 * Arguments :
 *	fp_t pT	Thread pointer for allocation
 *			int32u *pStack		Stack pointer
 * Return :
 *	int8s
 *		0	success
 *		1	fail
 *
 *********************************************************************************/

int32u 
Port_StackInit(fp_t pT, int32u *pStack)
{
  int32u *pStk;

  pStk = pStack;
#if 0
  *(--pStk) = (int32u)pT;		// PC
#endif
  *(--pStk) = (int32u)pT;		// PC
  *(--pStk) = (int32u)14;		// LR
  *(--pStk) = (int32u)12;		// R12
  *(--pStk) = (int32u)11;		// R11
  *(--pStk) = (int32u)10;		// R10
  *(--pStk) = (int32u)9;		// R09
  *(--pStk) = (int32u)8;		// R08
  *(--pStk) = (int32u)7;		// R07
  *(--pStk) = (int32u)6;		// R06
  *(--pStk) = (int32u)5;		// R05
  *(--pStk) = (int32u)4;		// R04
  *(--pStk) = (int32u)3;		// R03
  *(--pStk) = (int32u)2;		// R02
  *(--pStk) = (int32u)1;		// R01
  *(--pStk) = (int32u)0;		// R00
  *(--pStk) = (int32u)0x1f;		// CPSR

  return pStk;
}

static int32u *
Kern_StackAlloc(fp_t pT)
{
  int32u *pStack;
  pStack = (int32u *)MemAlloc(STACK_SIZE*sizeof(int32u));
  pStack = (int32u *)((int8s *)pStack + STACK_SIZE*sizeof(int32u));

  pStack = Port_StackInit(pT, pStack);
  return pStack;
}



tcb_t * 
Kern_ThreadCreate(int16u tid, fp_t pT, int16u prior,int16u deadLine )
{

#ifdef SCHED_DRR /* Dynamic Round Robin scheduling */
  tcb_t *newtcb;
  newtcb = (tcb_t *)MemAlloc(sizeof(tcb_t *));	// TCB creates
  Kern_ThreadInit(newtcb,tid,pT,prior,deadLine);		// TCB initializes and Stack allocates
  Kern_RdyListInsert(newtcb);							// Insert new TCB into Ready List

#endif //SCHED_DRR

//  Emul_ThreadCreate(pT,newtcb->tid);
  return newtcb;
}

tcb_t initThread;

void
Kern_KernelThreadCreate(){
  
  INIT_LIST_HEAD(&initThread.memList.allocLink);	  
  INIT_LIST_HEAD(&initThread.memList.memLink);	  
  
  /*In case scheduling policy changed, fields for other scheduling are 
	needed to set up */
  INIT_LIST_HEAD(&initThread.threadList);	  
  initThread.status = CREATED;
  pCurrThread_g = &initThread;	/*Set initThread as  current thread  */
  return ;
}




